try:
    from .models import SplitTestModel
except ImportError:
    from models import SplitTestModel
from builtins import str as text
import pickle
import terragon
import ast
import inspect
import tokenize
import types
import json
import sys
import re
import io
import os
import pprint as pp
from .reindenter import Reindenter
import logging


try:
    import dill
except:
    pass

def reindent(code):
    code = u"%s" % text(code)
    r = Reindenter(io.StringIO(code))
    r.run()
    out = io.StringIO()
    r.write(out)
    return out.getvalue()


def _in_directory(filepath, directory):
    #make both absolute
    directory = os.path.realpath(directory)
    filepath = os.path.realpath(filepath)
    #return true, if the common prefix of both is equal to directory
    #e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
    return os.path.commonprefix([filepath, directory]) == directory

def _is_on_syspath(filepath):
    if filepath is None:
        return False
    for libpath in sys.path:
        if libpath==os.getcwd():
            continue
        elif libpath!="":
            if _in_directory(filepath, libpath)==True:
                return True

    if os.environ.get("CONDA_DEFAULT_ENV"):
        # CONDA wierdness. check and see if CONDA is messing with our
        # syspath we're also going to have an exception for the yhat
        # library. not sure if this actually helps/hurts, but I don't
        # want to find out
        conda_env = os.environ["CONDA_DEFAULT_ENV"]
        conda_syspath = os.path.join(conda_env, "lib", "python2.7")
        if conda_syspath in filepath:
            return True

    if os.environ.get("VIRTUAL_ENV"):
        # same nonsense as above,but for virtualenv
        if _in_directory(filepath, os.environ["VIRTUAL_ENV"]):
            return True

    return False

def _get_source(func):
    """
    Gets the source of a function and handles cases when user is in an
    interactive session

    func: function
        a function who's source we want
    """
    # standardize source code indent to 4 spaces
    return reindent(_get_source_no_reindent(func))

def _get_source_no_reindent(func):
    try:
        return inspect.getsource(func)
    except:
        if inspect.isclass(func):
            source = ""
            source += "class %s(%s):" % (func.__name__, func.__base__.__name__)
            source += "\n"
            for name, method in inspect.getmembers(func, predicate=inspect.ismethod):
                if hasattr(method, '__wrapped_func__'):
                    wrapped_source = dill.source.getsource(method.__wrapped_func__)
                    wrapped_source = wrapped_source.split('\n')
                    nchar = len(wrapped_source[1]) - len(wrapped_source[1].lstrip())
                    wrapped_source[0] = " "*nchar + wrapped_source[0]
                    wrapped_source = "\n".join(wrapped_source)
                    wrapped_source = "\n" + reindent(wrapped_source) + "\n"
                    source += wrapped_source
                else:
                    source += reindent(inspect.getsource(method)) + "\n"
        else:
            return inspect.getsource(func) + "\n"
        return source


def _strip_function_source(src):
    """
    Takes the source code of a function and dedents it so that it.

    Parameters
    ----------
    src: string
        function code

    Returns
    -------
    source: string
        source code of the same function, but "dedented"
    """
    src = reindent(src)
    src = src.split('\n')
    n = len(src[0]) - len(src[0].lstrip())
    return "\n".join([line[n:] for line in src])

def _get_naked_loads(function):
    """
    Takes a reference to a function and determines which variables used in the
    function are not defined within the scope of the function.

    Parameters
    ----------
    function: function

    Returns
    -------
    variables: generator
        returns the variables in a function that are not:
            1) passed in as parameters
            2) created within the scope of the function
    """
    source = _get_source(function)
    source = _strip_function_source(source)
    logging.info("parsing source for %s\n" % str(function))
    logging.debug("%s\n" % source)
    tree = ast.parse(source)
    params = set()
    loaded = set()
    created = set()
    for thing in ast.walk(tree):
        thingvars = vars(thing)
        if "ctx" in thingvars:
            if isinstance(thingvars['ctx'], ast.Param):
                params.add(thingvars['id'])
            elif isinstance(thingvars['ctx'], ast.Load):
                if 'id' in thingvars:
                    loaded.add(thingvars['id'])
                    # variable = thingvars['id']
                    # if variable not in params and variable not in created:
                        # yield variable
            elif isinstance(thingvars['ctx'], ast.Store):
                if 'id' in thingvars:
                    created.add(thingvars['id'])

    for variable in loaded:
        # TODO: if created but loaded before that as something else
        if variable not in params and variable not in created:
            yield variable

def _extract_module(module_name, modules={}):
    logging.debug("\textracting module %s." % module_name)
    module = sys.modules.get(module_name)
    # check if we've already seen it
    if module in modules or module is None:
        logging.debug("\tmodule %s was None, returning." % module_name)
        return modules
    # make sure it's not a built in module
    elif hasattr(module, "__file__")==False:
        logging.debug("\tmodule %s did not have a `__file__` attribute, returning." % module_name)
        return modules
    elif _is_on_syspath(module.__file__)==False:
        logging.debug("\tfile being parsed is: %s\n" % module.__file__)

        if module.__file__.endswith(".py"):
            module_py = module.__file__
        elif module.__file__.endswith(".pyc"):
            module_py = module.__file__.replace(".pyc", ".py")
        else:
            logging.debug("\t%s is not a .py or .pyc skipping and returning: \n" % module.__file__)
            modules[module_name] = None
            return modules

        module_source = open(module_py, 'rb').read()
        parent_dir = module_py.replace(os.getcwd(), '').lstrip(os.sep)
        parent_dir = os.path.dirname(parent_dir)
        modules[module] = {
            "parent_dir": parent_dir,
            "name": os.path.basename(module_py),
            "source": module_source
        }
        # Try to extract init files
        finit = os.path.join(parent_dir, "__init__.py")
        isinit = os.path.isfile(finit)
        if isinit:
            init_source = open(finit, 'rb').read()
            modules['init'] = {
                "parent_dir": parent_dir,
                "name": "__init__.py",
                "source": init_source
            }

        logging.info("\tparsing source for %s\n" % module_name)
        logging.debug("%s\n" % module_source)

        logging.debug("\tparsing module source code for %s and looking for other modules to extract" % module_name)
        tree = ast.parse(module_source)
        for thing in ast.walk(tree):
            if hasattr(thing, "module"):
                modules.update(_extract_module(thing.module))
            elif isinstance(thing, (ast.Import, ast.ImportFrom)):
                for imp in thing.names:
                    if imp.name!="*":
                        modules.update(_extract_module(imp.name))
    else:
        modules[module_name] = None
    return modules

def _is_tensor(obj):
    try:
        return obj.__module__.startswith("tensorflow")
    except:
        try:
            return None != re.search( r'(tensorflow)', str(obj.__class__))
        except:
            return False

def _is_spark(obj):
    try:
        return obj.__module__.startswith("pyspark")
    except:
        try:
            return None != re.search( r'(pyspark)', str(obj.__class__))
        except:
            return False

def _spider_function(function, session, pickles={}):
    """
    Takes a function and global variables referenced in an environment and
    recursively finds dependencies required in order to execute the function.
    This includes references to classes, libraries, variables, functions, etc.

    Parameters
    ----------
    function: function
        a function referenced in an environment
    session: dictionary
        variables referenced from a seperate environment; i.e. globals()
    pickles: dictionary
        holds the variables needed to execute the function

    Returns
    -------
    imports: list
        list of import statements required for the function to execute
    source: string
        source code of the function
    pickles: dictionary
        dictionary of variable names and their values as pickled strings
    """

    if '_objects_seen' not in pickles:
        pickles['_objects_seen'] = []
    pickles['_objects_seen'].append(str(function))
    imports = []
    modules = {}
    source = "# code for %s\n" % (str(function))
    if isinstance(function, types.ModuleType):
        logging.debug("object (%s) is a `types.ModuleType`. skipping _get_source" % str(function))
        pass
    else:
        source += _get_source(function) + '\n'

    logging.debug("finding objects that need to be serialized")
    for varname in _get_naked_loads(function):
        logging.debug("\t%s" % varname)
        if varname in pickles['_objects_seen']:
            logging.debug("\t%s has already been processed. skipping." % varname)
            continue
        pickles['_objects_seen'].append(varname)
        if varname not in session:
            logging.debug("\t%s not found in session. skipping." % varname)
            continue
        obj = session[varname]
        # checking to see if this is an instance of an object
        if hasattr(obj, "__name__")==False:
            logging.debug("\t%s has attribute __name__, performing serialization checks." % varname)
            if _is_tensor(obj):
                logging.debug("\t%s is from tensorflow skipping." % varname)
                continue
            elif _is_spark(obj):
                logging.debug("\t%s is from spark. serializing using `dumps_spark_to_base64`." % varname)
                pickles[varname] = terragon.dumps_spark_to_base64(session['sc'], obj)
            else:
                logging.debug("\tno special serialization requirement detected for %s. using dumps_to_base64." % varname)
                pickles[varname] = terragon.dumps_to_base64(obj)

        if hasattr(obj, "__module__"):
            logging.debug("\tobject %s has attribute __module__" % varname)
            if obj.__module__=="__main__":
                logging.debug("\t%s is in module __main__" % varname)
                new_imports, new_source, new_pickles, new_modules = _spider_function(obj, session, pickles)
                source += new_source + '\n'
                imports += new_imports
                pickles.update(new_pickles)
                modules.update(new_modules)
            else:
                logging.debug("\t%s is a submodule (not __main__). extracting/saving source code and import statements." % varname)
                modules.update(_extract_module(obj.__module__))
                ref = inspect.getmodule(obj).__name__
                if hasattr(obj, "func_name") and obj.__name__!=varname:
                    imports.append("from %s import %s as %s" % (ref, obj.__name__, varname))
                else:
                    # we need to figure out how to import this library. i'm not
                    # sure exactly what the right way to get the module and
                    # class name, but this works just fine
                    try:
                        import_statement = "from %s import %s" % (ref, varname)
                        exec(import_statement, locals())
                        imports.append(import_statement)
                    except:
                        try:
                            import_statement = "from %s import %s" % (ref, obj.__class__.__name__)
                            exec(import_statement, locals())
                            imports.append(import_statement)
                        except:
                            pass

        elif isinstance(obj, types.ModuleType):
            logging.debug("\tobject %s is a `types.ModuleType`, extracting/saving source code and import statements." % varname)
            modules.update(_extract_module(obj.__name__))
            if obj.__name__!=varname:
                imports.append("import %s as %s" % (obj.__name__, varname))
            else:
                imports.append("import %s" % (varname))
        else:
            logging.debug("\tno special cases detected for %s. pickling using dumps_to_base64" % varname)
            # catch all. if all else fails, pickle it
            pickles[varname] = terragon.dumps_to_base64(obj)

    return imports, source, pickles, modules

def _detect_future_imports(session):
    """
    Detect all __future__ imports. Since these imports must come first in
    the source code, these are attempt to be detected outside of other
    spidering.

    Parameters
    ----------
    session: dictionary
        globals() from the user's environment
    """
    imports = []
    for k in list(session.keys()):
        v = session[k]
        if hasattr(v,"__module__"):
            if v.__module__ == "__future__":
                if k is 'print_function':
                    continue
                if not k.startswith('_'):
                    imports.append("from __future__ import %s" % k)
    return imports

def save_function(function, session):
    """
    Saves a user's session and all dependencies to a big 'ole JSON object with
    accompanying pickles for any variable.

    Parameters
    ----------
    function: function
        function we're saving
    session: dictionary
        globals() from the user's environment
    """

    future_imports = "\n".join(_detect_future_imports(session))
    logging.debug("Detecting future import for session. Found %s" % future_imports)

    logging.debug("Spidering model class %s" % str(function))
    imports, source_code, pickles, modules = _spider_function(function, session)

    # de-dup and order the imports
    imports = sorted(list(set(imports)))

    imports.append("import random") # we need this for the execute method for a SplitTestModel
    imports.append("import json")
    imports.append("import pickle")
    imports.append("import terragon")
    source_code = "\n".join(imports) + "\n\n\n" + source_code
    pickModules = []

    logging.debug("%d sub-modules found. encoding source code as utf-8." % len(modules))
    for _, value in list(modules.items()):
        if value is not None:
            try:
                value['source'] = value['source'].decode(encoding="utf-8")
                pickModules.append(value)
            except Exception as e:
                logging.debug("could not encode source for %s: %s" % (str(value), str(e)))
                pickModules.append(value)

    pickles = {
        "objects": pickles,
        "future": future_imports,
        "code": source_code,
        "modules": pickModules
    }

    logging.debug("Future Imports Found: %s" % str(future_imports))
    logging.debug("Objects Found: %s" % str(pickles.keys()))
    logging.debug("Source Code Found:\n%s" % source_code)
    logging.debug("Modules Found:\n%s" % pp.pformat(pickModules, indent=2))

    if "_objects_seen" in pickles["objects"]:
        del pickles["objects"]["_objects_seen"]
    return pickles



# models
from .deployment.models import YhatModel
from .deployment.models import SplitTestModel, Variant
# preprocessing decorator
from .deployment.input_and_output import preprocess
from .deployment.input_and_output import df_to_json
# client
from .api import Yhat
from .version import __version__
from .yhat_json import json_dumps



from flask import Flask
from flask import request
app = Flask(__name__)

@app.route('/verify', methods=["POST"])
def verify():
    username = request.args.get('username', '')
    apikey = request.args.get('apikey', '')
    print('username: ' + username)
    print('apikey: ' + apikey)
    return "{\"success\": \"true\"}", 200

@app.route('/<path:path>', methods=["GET", "POST"])
def catch_all(path):
    print(request)
    return "You send a request to: %s\n" % path, 200

if __name__ == '__main__':
    app.run()



import inspect
import json
import pickle
import sys
import os
import warnings
import random
import re

from .input_and_output import df_to_df, parse_json, preprocess

try:
    import pandas as pd
except:
    warnings.warn("Could not import pandas")


class YhatModel(object):
    """
    A simple class structure for your model. Place all of the code you want to
    execute in the `execute` method.

    Parameters
    ----------
    filename: string
        filename of saved model
    """
    def __init__(self, filename=None):
        if filename is not None:
            pickles = json.load(open(filename, 'rb'))
            for varname, pickled_value in pickles.get('objects', {}).items():
                globals()[varname] = pickle.loads(pickled_value)

    @preprocess
    def execute(self, data):
        """
        This is the execution plan for your model. It defines what routines you
        will execute and how the input and output will be handled.

        Parameters
        ----------
        data: dict or DataFrame
            the datatype is decided by the IO Specification; (df_to_df,
            dict_to_dict, etc.).
        """
        pass

class Variant(dict):
    """
    A Variant is an option to use in a SplitTestModel. Variants are composed of 3 properties:

        - label: string, this will be included in the model output to indicate which variant was run in the split test
        - method: string, this is a reference to the method that will be used when this variant is used
        - traffic_allocation: float, the percentage of traffic that will be routed to this variant for the split test

    >>> variants = [
    ...    Variant("A", "execute_a", 0.5),
    ...    Variant("B", "execute_b", 0.5),
    ... ]
    >>> variants = [
    ...    Variant("A", "execute_a", 0.5),
    ...    Variant("B", "execute_b", 0.3),
    ...    Variant("C", "execute_c", 0.2),
    ... ]
    >>> variants = [
    ...    Variant("SVM", "execute_svc", 0.7),
    ...    Variant("Logistic", "execute_logit", 0.3)
    ... ]
    """
    def __init__(self, label=None, method=None, traffic_allocation=0.5):
        self['label'] = label
        self['method'] = method
        self['traffic_allocation'] = traffic_allocation


class SplitTestModel(YhatModel):
    """
    Create an A/B testable model. This will split traffic between different
    models (specified as methods in your class).

    Examples
    ========

    class YetAnotherModel(SplitTestModel):
        def setup_variants(self):
            return [
                Variant(label="SVC", method="execute_svc", traffic_allocation=0.5),
                Variant(label="LOGIT", method="execute_logit", traffic_allocation=0.5)
            ]
    yam = YetAnotherModel()
    yam.execute({ "n": 1 })
    # {"variant": "SVC", "probability": 0.45}

    class MyModel(SplitTestModel):
        def setup_variants(self):
            return [
                Variant("A", "execute_a", 0.5),
                Variant("B", "execute_b", 0.4),
                Variant("C", "execute_c", 0.1)
            ]
    model = MyModel()
    model.execute({ "n": 1 })
    # {"variant": "C", "probability": 0.12}

    class AnotherModel(SplitTestModel):
        def setup_variants(self):
            return [
                Variant("SVC", "execute_svc", 0.5),
                Variant("LOGIT", "execute_logit", 0.4),
                Variant("DEFAULT", "execute_default", 0.1)
            ]
    am = AnotherModel()
    am.execute({ "n": 1 })
    # {"variant": "LOGIT", "probability": 0.32}
    """

    def __init__(self, variants=None):
        if variants is None:
            if hasattr(self, "variants"):
                variants = self.variants
            elif hasattr(self, "setup_variants"):
                variants = self.setup_variants()

        if sum(v['traffic_allocation'] for v in variants) != 1:
            msg = " + ".join([str(v['traffic_allocation']) for v in variants])
            msg += " = " + str(sum(v['traffic_allocation'] for v in variants))
            raise Exception("Your total traffic allocation for all variants must sum to 1!\n\t%s" % msg)

        for variant in variants:
            try:
                getattr(self, variant['method'])
            except:
                raise Exception("Method '{}' for variant '{}' does not exist".format(variant['method'], variant['label']))

        self.variants = variants
        super(SplitTestModel, self).__init__()


    def execute(self, data):
        r = random.random()
        total = 0.0
        for variant in self.variants:
            total += variant['traffic_allocation']
            if r < total:
                method = getattr(self, variant['method'])
                output = method(data)
                output['variant'] = variant['label']
                return output

class Model(object):
    def __init__(self, **kwargs):
        for kw, arg in kwargs.iteritems():
            if kw=="requirements":
                arg = open(arg).read().strip()
            setattr(self, kw, arg)



import warnings
import json

try:
    import pandas as pd
except:
    warnings.warn("Could not import pandas")

def make_df(data):
    """
    Takes an arbitrary data structure and tries to convert it into a data frame

    Parameters
    ----------
    data: object
        data to be converted into data frame

    Returns
    -------
    data: data frame
    """
    try:
        data = pd.DataFrame(data)
    except:
        data = pd.DataFrame([data])
    return data


def parse_json(somejson):
    """
    Attempts to parse JSON and return a serialized object

    Parameters
    ----------
    somejson: object

    Returns
    -------
    object: dictionary or list
    """
    try:
        return json.loads(somejson)
    except Exception as e:
        print("Input was not valid JSON.")
        print("==> %s" % str(e))

class df_to_df(object):
    """
    Decorator that coreces input and outputs into data frames.
    """
    def __init__(self, func):
        self.func = func
    
    def __call__(self, *args):
        data = args[0]
        data = make_df(data)
        result = self.func(self, data)
        return make_df(result)

class df_to_dict(object):
    """
    Decorator that coreces input to a data frame and requires the output to be
    a dictionary.
    """
    def __init__(self, func):
        self.func = func
    
    def __call__(self, *args):
        data = args[0]
        data = make_df(data)
        result = self.func(self, data)
        if isinstance(result, dict):
            return result
        else:
            msg = "Dictionary was not returned. '%s' was returned instead"
            msg = msg % str(type(result))
            raise Exception(msg)


class dict_to_dict(object):
    """
    Decorator that needs inputs and outputs to be dictionaries.
    """
    def __init__(self, func):
        self.func = func
    
    def __call__(self, *args):
        data = args[0]
        if isinstance(result, dict)==False:
            msg = "Dictionary was not given. '%s' was input instead"
            msg = msg % str(type(result))
            raise Exception(msg)
        result = self.func(self, data)
        if isinstance(result, dict):
            return result
        else:
            msg = "Dictionary was not returned. '%s' was returned instead"
            msg = msg % str(type(result))
            raise Exception(msg)


def preprocess(func=None, **options):
    """
    Decorator for defining the following:
        1) data type for incoming data
        2) data type for returned data
        3) how to handle null values

    Parameters
    ----------
    in_type: (optional), dictionary or data frame
        indicates what the input data type should be
    out_type: (optional), dictionary or data frame
        indicates what the returned data type should be

    Returns
    -------
    partial/inner: function
        a function
    """
    if func != None:
        # We received the function on this call, so we can define
        # and return the inner function
        def inner(*args, **kwargs):        
            in_type = options.get('in_type', pd.DataFrame)
            out_type = options.get('out_type', pd.DataFrame)
            data = args[1]
            if in_type==pd.DataFrame:
                data = make_df(data)
            data = func(args[0], data)
            if out_type==pd.DataFrame:
                data = make_df(data)
            return data
        inner.__wrapped_func__ = func
   
        return inner
  
    else:
        # We didn't receive the function on this call, so the return value
        # of this call will receive it, and we're getting the options now.
        def partial_inner(func):
            return preprocess(func, **options)   
        return partial_inner

def df_to_json(df):
    """
    Convert Pandas DataFrame to Yhat compatable json

    Parameters
    ----------
    df: Pandas DataFrame

    Returns
    -------
    json string: str
    """
    try:
        pd
    except NameError:
        raise ImportError("df_to_json() requires pandas")
    if not isinstance(df,pd.DataFrame):
        raise ValueError("'df' parameter must be a pandas DataFrame")
    if df.index.name:
        msg = "index values are NOT maintained through jsonifification," + \
                " consider resetting index";
        warnings.warn(msg)
    df_values = df.transpose().to_json(orient='values',date_format='iso')
    df_values = json.loads(df_values)
    try:
        from collections import OrderedDict
        df_values = OrderedDict(list(zip(df.columns,df_values)))
    except ImportError:
        df_values = dict(list(zip(df.columns,df_values)))
    return json.dumps(df_values)



import unittest
import numpy as np
from yhat import json_dumps
import json

class TestYhatJson(unittest.TestCase):

    def test_dumps_dict(self):
        x = { "x": 1 }
        self.assertEqual(json.dumps(x), json_dumps(x))

    def test_dumps_list(self):
        x = [ "a", "b", "c" ]
        self.assertEqual(json.dumps(x), json_dumps(x))

    def test_dumps_string(self):
        x = [ "a", "b", "c" ]
        self.assertEqual(json.dumps(x), json_dumps(x))

    def test_dumps_int(self):
        x = 10
        self.assertEqual(json.dumps(x), json_dumps(x))

    def test_dumps_float(self):
        x = 10.1
        self.assertEqual(json.dumps(x), json_dumps(x))

    def test_dumps_nan(self):
        x = np.nan
        self.assertEqual(json_dumps(x), 'null')

    def test_dumps_nan_list(self):
        x = [np.nan, np.nan, np.nan]
        self.assertEqual(json_dumps(x), '[null, null, null]')

    def test_dumps_numpy_array(self):
        x = np.arange(100)
        self.assertEqual(json_dumps(x), json.dumps(list(range(100))))

if __name__=="__main__":
    unittest.main()



import json
# we're not going to require numpy as a hard requirement as it can be a huge
# pain to install. *most* of our customers will already have it installed. the ones
# that don't are most likely just doing a lightweight test (i.e. HelloWorld) and
# installing numpy would just be really annoying.
try:
    import numpy as np
except ImportError as e:
    np = None

class NumpyAwareJSONEncoder(json.JSONEncoder):
    """
    NumpyAwareJSONEncoder makes numpy arrays JSON serializeable. This is important
    because most mathematical libraries return numpy arrays instead of python lists.

    If there was a new numpy data type that you wanted to make JSON serializeable,
    all you'd need to do is edit the `default` method below. This could also be
    modified in the future to serialize non-numpy specific data types (though you'd
    prboably want to change the name of the class).
    """
    def __init__(self, nan_str="null", **kwargs):
        super(NumpyAwareJSONEncoder, self).__init__(**kwargs)
        self.nan_str = nan_str
        self.allow_nan = True

    def default(self, obj):
        if np and isinstance(obj, np.ndarray) and obj.ndim == 1:
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

    # uses code from official python json.encoder module. Same licence applies.
    def iterencode(self, o, _one_shot=False):
        """
        Encode the given object and yield each string representation as
        available.

        For example:
            for chunk in JSONEncoder().iterencode(bigobject):
                mysocket.write(chunk)
        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.encode_basestring_ascii

        else:
            _encoder = json.encoder.encode_basestring
        # if self.encoding != 'utf-8':
        #     def _encoder(o, _orig_encoder=_encoder, _encoding=self.encoding):
        #         if isinstance(o, str):
        #             o = o.decode(_encoding)
        #         return _orig_encoder(o)

        def floatstr(o, allow_nan=self.allow_nan, _repr=repr,
            _inf=json.encoder.INFINITY, _neginf=-json.encoder.INFINITY,
            nan_str=self.nan_str):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if o != o:
                text = nan_str
            elif o == _inf:
                text = 'Infinity'
            elif o == _neginf:
                text = '-Infinity'
            else:
                return _repr(o)

            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " +
                    repr(o))

            return text

        _iterencode = json.encoder._make_iterencode(
                markers, self.default, _encoder, self.indent, floatstr,
                self.key_separator, self.item_separator, self.sort_keys,
                self.skipkeys, _one_shot)
        return _iterencode(o, 0)


# we're going to create a json de-serializer that will by default use
# NumpyAwareJSONEncoder with json
def json_dumps(data, **kwargs):
    """
    Uses json.dumps to serialize data into JSON. In addition to the standard
    json.dumps function, we're also using the NumpyAwareJSONEncoder to handle
    numpy arrays and the `ignore_nan` parameter by default to handle np.nan values.
    """
    return json.dumps(data, cls=NumpyAwareJSONEncoder, allow_nan=False)



import os

from distutils.core import setup
from setuptools import find_packages

# Get version from defined python file
with open(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'yhat',
        'version.py'
    )
) as fh:
    version = fh.read().strip()
exec(version)

setup(
    name="yhat",
    version=__version__,
    author="Greg Lamp",
    author_email="greg@yhathq.com",
    url="https://github.com/yhat/yhat-client",
    packages=find_packages(),
    description="Python client for Yhat (http://yhat.com/)",
    license="BSD",
    classifiers=(
    ),
    install_requires=[
        "pip",
        "future==0.15.2",
        "dill==0.2.5",
        "terragon==0.3.0",
        "progressbar2==3.10.1",
        "requests==2.11.1",
        "requests-toolbelt==0.7.0"
    ],
    long_description=open("README.rst").read(),
    keywords=['yhat', 'scikits', 'numpy', 'pandas'],
)



from pip._vendor.pkg_resources import Requirement, RequirementParseError
try:
    from pip.utils import get_installed_distributions # pip 6.0
except ImportError:
    from pip.util import get_installed_distributions  # pip 1.5.x
from warnings import warn
import types


"""
This package handles the requirments portion of the yhat client.
If the user has set the autodetect flag to False in the deploy commpand, then
we only run getExplicitRequirmets, otherwise we will run getImplicitRequirements.
The implicit piece attempts to pull all the Python library imports and their
version numbers from a session's globals. It will then merge these with the user
defined requirements.
The final steps are to print the requirments and return a one line string
that can be put into the bundle
"""

def _get_package_name(obj):
    """Returns the package name (e.g. "sklearn") for a Python object"""
    try:
        if isinstance(obj, types.ModuleType):
            return obj.__package__.split(".")[0]
        elif isinstance(obj, type):
            return obj.__module__.split(".")[0]
        elif isinstance(obj, object):
            return obj.__class__.__module__.split(".")[0]
        else:
            return None
    except:
        return None

def parseUserRequirementsList(reqList):
    PACKAGE_LIMIT = 25
    pkgCount = 0
    userReqsRaw = []
    userReqs = []
    for r in reqList:
        # Look for .txt
        r = r.strip().strip('\n')
        if r[-4:] == '.txt':
            pkgList = []
            f = open(r, 'r')
            for line in f:
                line = line.strip('\n').strip()
                if line[0] != '#':
                    pkgList.append(line)
            f.close()
            reqList.extend(pkgList)
        else:
            try:
                if r[:4] != 'yhat':
                    userReqsRaw.append(Requirement.parse(r))
                    pkgCount += 1
            except RequirementParseError:
                if r[:3] == 'git':
                    userReqsRaw.append(r)
                    pkgCount += 1
                else:
                    print('Package ' + r + ' was not recognized as a valid package.')
            except:
                print("Unexpected error:", sys.exc_info()[0])
                raise

    if pkgCount > PACKAGE_LIMIT:
        warn(
            "\nYou have tried to add %s requirmements, which exceeds the maximum amount you can add during a deployment.\n"
            "If possible, please consider explicily adding fewer requirements and try to re-deploy.\n"
            "Or if you require this many packages, contact Yhat to upgrade your base image.\n" % str(pkgCount)
        )
        # block the deployment
        sys.exit()
    else:
        for pkg in userReqsRaw:
            if pkg not in userReqs:
                userReqs.append(pkg)

    return userReqs

def initializeRequirements(model):
    requirements = {
        'modelSpecified': [],
        'required': [],
        'autodetected': []
    }
    userReqs = getattr(model, "REQUIREMENTS", "")
    if isinstance(userReqs, str):
        userReqs = [r for r in userReqs.splitlines() if r]
    if userReqs:
        userReqs = parseUserRequirementsList(userReqs)
        requirements['modelSpecified'] = userReqs

    # Always add yhat package to required list with the installed version.
    import yhat
    yhatReq = Requirement.parse('yhat==%s' % yhat.__version__)
    requirements['required'].append(yhatReq)

    return requirements

def getImplicitRequirements(model, session):
    requirements = initializeRequirements(model)
    requirements = implicit(session, requirements)
    return bundleRequirments(requirements)

def getExplicitRequirements(model, session):
    requirements = initializeRequirements(model)
    return bundleRequirments(requirements)

def printRequirements(requirements):
    for cat, reqList in list(requirements.items()):
        if reqList:
            if cat == "required":
                print("required packages")
            elif cat == "modelSpecified":
                print("model specified requirements")
            elif cat == "autodetected":
                print("autodetected packages")
            for r in reqList:
                if "==" not in str(r) and str(r)[:3] != 'git':
                    r = str(r) + " (latest)"
                print(" [+] " + str(r))

def bundleRequirments(requirements):
    # Put the requirements into a structure for the bundle
    reqList = []
    mergedReqs = merge(requirements)
    for reqs in mergedReqs.values():
        if reqs:
            for r in reqs:
                reqList.append(r)
    bundleString = "\n".join(
        str(r) for r in reqList
    )
    printRequirements(requirements)

    return bundleString

def implicit(session, requirements):
    """
    Returns a list of Requirement instances for all the library dependencies
    of a given session. These are matched using the contents of "top_level.txt"
    metadata for all package names in the session.
    """
    package_names = [_get_package_name(g) for g in list(session.values())]
    package_names = set([_f for _f in package_names if _f])

    reqs = {}
    for d in get_installed_distributions():
        for top_level in d._get_metadata("top_level.txt"):
            if top_level in package_names and top_level != 'yhat':
                # Sanity check: if a distribution is already in our
                # requirements, make sure we only keep the latest version.
                if d.project_name in reqs:
                    reqs[d.project_name] = max(reqs[d.project_name], d.version)
                else:
                    reqs[d.project_name] = d.version

    requirements['autodetected'] = [Requirement.parse('%s==%s' % r) for r in list(reqs.items())]
    return requirements

def merge(requirements):
    """
    Merges autodetected and explicit requirements together. Autodetected
    requirements are pulled out the user's session (i.e. globals()).
    Explicit requirements are provided directly by the user. This
    function reconciles them and merges them into one set of requirements.
    Warnings are given to the user in case of version mismatch.
    Because we want to move away from implicitly getting requirements, we warn
    the user if there are implicitly detecdet but not explicitly stated
    requirements.
    """
    implicit_dict = {}
    for r in requirements['autodetected']:
        implicit_dict[r.project_name] = r

    explicit_dict = {}
    for r in requirements['modelSpecified']:
        if type(r) != str:
            explicit_dict[r.project_name] = r

    for project_name, exp_req in list(explicit_dict.items()):
        # To be polite, we keep the explicit dependencies and add the implicit
        # ones to them. We respect versions on the former, except in the case
        # of yhat, which should be the installed version.
        if project_name in implicit_dict:
            imp_req = implicit_dict[project_name]
            if exp_req == imp_req:
                # we only need one of these, remove the implicit, but don't need
                # to warn the user
                requirements['autodetected'].remove(Requirement.parse(str(imp_req)))
            elif project_name == "yhat":
                warn(
                    "The yhat package can be removed from REQUIREMENTS. "
                    "It is required and added for you."
                )
                try:
                    requirements['autodetected'].remove(Requirement.parse(str(imp_req)))
                    requirements['modelSpecified'].remove(Requirement.parse(str(exp_req)))
                except:
                    pass
            else:
                warn(
                    "You have explicitly %s as a requirement, which differs from %s, "
                    "which is was implicitly found to be installed locally\n"
                    "Deploying with explicitly defined package: %s " % (exp_req, imp_req, exp_req)
                )
                requirements['autodetected'].remove(Requirement.parse(str(imp_req)))

    # Loop through the implicit dict and notify users if they haven't explicitly
    # specified a requirement
    for project_name, imp_req in list(implicit_dict.items()):
        if project_name not in explicit_dict:
            warn(
                "Dependency %s was found with autodetection, but we recommend "
                "explicitly stating your requirements." % (imp_req)
            )

    return requirements
